package aud.exam.prep.tree;

import java.util.Comparator;

/**
 * Instances of this interface can work with any "valid" tree of type <code>T</code>.
 * For what is valid, see {@link #check(T, Comparator)}.
 * For this processor trees always have to be "in order".
 * This is a literal translation of the "inspiration sheet" into java.
 *
 * @param <V> Type of elements stored in <code>T</code>
 * @param <T> Type of the tree
 */
public interface OrderedTreeProcessor<V, T> {

    boolean find(T t, V v, Comparator<V> cmp);

    boolean override(T t, V from, V to, Comparator<V> cmp);

    boolean overrideAll(T t, V from, V to, Comparator<V> cmp);

    T insert(T t, V v, Comparator<V> cmp);

    boolean remove(T t, V v, Comparator<V> cmp);

    boolean removeAll(T t, V v, Comparator<V> cmp);

    V max(T t, Comparator<V> cmp);

    V secondMax(T t, Comparator<V> cmp);

    int height(T t);

    boolean isBalanced(T t);

    int numberOfNodes(T t);

    int numberOfNodesOnLevel(T t, int level);

    T rightmostNodeInLeftSubtree(T t);

    T leftmostNodeInRightSubtree(T t);

    T invert(T t);

    T clone(T t);

    /**
     * Check whether ot not the given tree is "valid" according to
     * what is considered valid depends on the specific data structure.
     * This is also used by the tests to verify implementations.
     * This has to be tested separate for each data structure.
     *
     * @param t A tree
     * @return true iff the given tree is valid
     */
    boolean check(T t, Comparator<V> cmp);

    /**
     * Create an empty tree.
     *
     * @return A new, empty tree
     */
    T newEmptyTree();

    /**
     * Return an {@link Iterable}, or a lambda returning {@link java.util.Iterator},
     * over the elements of the tree <code>t</code>.
     * This is required by the tests.
     *
     * @param t A tree
     * @return An {@link Iterable} over t
     */
    Iterable<V> iterate(T t);

    /**
     * Print a tree to standard out
     *
     * @param t A tree
     */
    void print(T t);
}